@@ -1,5 +1,5 @@
 /* kcolor.h - Color-oriented function declarations for gifsicle.
-   Copyright (C) 2013-2014 Eddie Kohler, ekohler@gmail.com
+   Copyright (C) 2013-2017 Eddie Kohler, ekohler@gmail.com
    This file is part of gifsicle.
 
    Gifsicle is free software. It is distributed under the GNU Public License,
@@ -25,11 +25,19 @@ typedef struct kcolor {
     int16_t a[3];
 } kcolor;
 
-#define KC_CLAMPV(v) ((v) < 0 ? 0 : ((v) < KC_MAX ? (v) : KC_MAX))
+#undef min
+#undef max
+#define min(a, b)       ((a) < (b) ? (a) : (b))
+#define max(a, b)       ((a) > (b) ? (a) : (b))
+
+#define KC_CLAMPV(v) (max(0, min((v), KC_MAX)))
 
 typedef union kacolor {
     kcolor k;
     int16_t a[4];
+#if HAVE_INT64_T
+    int64_t q; /* to get better alignment */
+#endif
 } kacolor;
 
 
@@ -38,26 +46,17 @@ typedef union kacolor {
 extern uint16_t* gamma_tables[2];
 
 
-/* set `*x` to the gamma transformation of `a0/a1/a2` [RGB] */
-static inline void kc_set8g(kcolor* x, int a0, int a1, int a2) {
-    x->a[0] = gamma_tables[0][a0];
-    x->a[1] = gamma_tables[0][a1];
-    x->a[2] = gamma_tables[0][a2];
+/* set `*kc` to the gamma transformation of `a0/a1/a2` [RGB] */
+static inline void kc_set8g(kcolor* kc, int a0, int a1, int a2) {
+    kc->a[0] = gamma_tables[0][a0];
+    kc->a[1] = gamma_tables[0][a1];
+    kc->a[2] = gamma_tables[0][a2];
 }
 
 /* return the gamma transformation of `a0/a1/a2` [RGB] */
 static inline kcolor kc_make8g(int a0, int a1, int a2) {
-    kcolor x;
-    kc_set8g(&x, a0, a1, a2);
-    return x;
-}
-
-/* return the kcolor representation of `*gfc` (no gamma transformation) */
-static inline kcolor kc_makegfc(const Gif_Color* gfc) {
     kcolor kc;
-    kc.a[0] = (gfc->gfc_red << 8) + gfc->gfc_red;
-    kc.a[1] = (gfc->gfc_green << 8) + gfc->gfc_green;
-    kc.a[2] = (gfc->gfc_blue << 8) + gfc->gfc_blue;
+    kc_set8g(&kc, a0, a1, a2);
     return kc;
 }
 
@@ -66,6 +65,20 @@ static inline kcolor kc_makegfcg(const Gif_Color* gfc) {
     return kc_make8g(gfc->gfc_red, gfc->gfc_green, gfc->gfc_blue);
 }
 
+/* return the uncorrected representation of `a0/a1/a2` [RGB] */
+static inline kcolor kc_make8ng(int a0, int a1, int a2) {
+    kcolor kc;
+    kc.a[0] = (a0 << 7) + (a0 >> 1);
+    kc.a[1] = (a1 << 7) + (a1 >> 1);
+    kc.a[2] = (a2 << 7) + (a2 >> 1);
+    return kc;
+}
+
+/* return the kcolor representation of `*gfc` (no gamma transformation) */
+static inline kcolor kc_makegfcng(const Gif_Color* gfc) {
+    return kc_make8ng(gfc->gfc_red, gfc->gfc_green, gfc->gfc_blue);
+}
+
 /* return transparency */
 static inline kacolor kac_transparent() {
     kacolor x;
@@ -101,17 +114,20 @@ static inline uint32_t kc_distance(const kcolor* x, const kcolor* y) {
 
 /* return the luminance value for `*x`; result is between 0 and KC_MAX */
 static inline int kc_luminance(const kcolor* x) {
-    return (306 * x->a[0] + 601 * x->a[1] + 117 * x->a[2]) >> 10;
+    return (55 * x->a[0] + 183 * x->a[1] + 19 * x->a[2]) >> 8;
 }
 
 /* set `*x` to the grayscale version of `*x`, transformed by luminance */
 static inline void kc_luminance_transform(kcolor* x) {
     /* For grayscale colormaps, use distance in luminance space instead of
+
        distance in RGB space. The weights for the R,G,B components in
-       luminance space are 0.299,0.587,0.114. Using the proportional factors
-       306, 601, and 117 we get a scaled gray value between 0 and 255 *
-       1024. Thanks to Christian Kumpf, <kumpf@igd.fhg.de>, for providing a
-       patch. */
+       luminance space are 0.2126,0.7152,0.0722. (That's ITU primaries, which
+       are compatible with sRGB; NTSC recommended our previous values,
+       0.299,0.587,0.114.) Using the proportional factors 55,183,19 we get a
+       scaled gray value between 0 and 255 * 257; dividing by 256 gives us
+       what we want. Thanks to Christian Kumpf, <kumpf@igd.fhg.de>, for
+       providing a patch.*/
     x->a[0] = x->a[1] = x->a[2] = kc_luminance(x);
 }
 
@@ -234,12 +250,60 @@ Gif_Colormap* colormap_flat_diversity(kchist* kch, Gt_OutputData* od);
 Gif_Colormap* colormap_median_cut(kchist* kch, Gt_OutputData* od);
 
 
-typedef struct scale_color {
-    double a[4];
+#if HAVE_SIMD && HAVE_VECTOR_SIZE_VECTOR_TYPES
+typedef float float4 __attribute__((vector_size (sizeof(float) * 4)));
+typedef int int4 __attribute__((vector_size (sizeof(int) * 4)));
+#elif HAVE_SIMD && HAVE_EXT_VECTOR_TYPE_VECTOR_TYPES
+typedef float float4 __attribute__((ext_vector_type (4)));
+#else
+typedef float float4[4];
+#endif
+
+typedef union scale_color {
+    float4 a;
 } scale_color;
 
 static inline void sc_clear(scale_color* x) {
     x->a[0] = x->a[1] = x->a[2] = x->a[3] = 0;
 }
 
+static inline scale_color sc_makekc(const kcolor* k) {
+    scale_color sc;
+    sc.a[0] = k->a[0];
+    sc.a[1] = k->a[1];
+    sc.a[2] = k->a[2];
+    sc.a[3] = KC_MAX;
+    return sc;
+}
+
+static inline scale_color sc_make(float a0, float a1, float a2, float a3) {
+    scale_color sc;
+    sc.a[0] = a0;
+    sc.a[1] = a1;
+    sc.a[2] = a2;
+    sc.a[3] = a3;
+    return sc;
+}
+
+#if HAVE_SIMD
+# define SCVEC_ADDV(sc, sc2) (sc).a += (sc2).a
+# define SCVEC_MULV(sc, sc2) (sc).a *= (sc2).a
+# define SCVEC_MULF(sc, f) (sc).a *= (f)
+# define SCVEC_DIVF(sc, f) (sc).a /= (f)
+# define SCVEC_ADDVxF(sc, sc2, f) (sc).a += (sc2).a * (f)
+# if HAVE___BUILTIN_SHUFFLEVECTOR
+#  define SCVEC_ROT3(out, sc) do { (out).a = __builtin_shufflevector((sc).a, (sc).a, 1, 2, 0, 3); } while (0)
+# else
+#  define SCVEC_ROT3(out, sc) do { int4 shufmask__ = {1, 2, 0, 3}; (out).a = __builtin_shuffle((sc).a, shufmask__); } while (0)
+# endif
+#else
+# define SCVEC_FOREACH(t) do { int k__; for (k__ = 0; k__ != 4; ++k__) { t; } } while (0)
+# define SCVEC_ADDV(sc, sc2) SCVEC_FOREACH((sc).a[k__] += (sc2).a[k__])
+# define SCVEC_MULV(sc, sc2) SCVEC_FOREACH((sc).a[k__] *= (sc2).a[k__])
+# define SCVEC_MULF(sc, f) SCVEC_FOREACH((sc).a[k__] *= (f))
+# define SCVEC_DIVF(sc, f) SCVEC_FOREACH((sc).a[k__] /= (f))
+# define SCVEC_ADDVxF(sc, sc2, f) SCVEC_FOREACH((sc).a[k__] += (sc2).a[k__] * (f))
+# define SCVEC_ROT3(out, sc) do { float __a0 = (sc).a[0]; (out).a[0] = (sc).a[1]; (out).a[1] = (sc).a[2]; (out).a[2] = __a0; (out).a[3] = (sc).a[3]; } while (0)
+#endif
+
 #endif
